# Standard albatross tags
#
import string
import time
import template

# Empty content object which tags can use as default value for
# optional child tag content.  <al-head>, <al-tail>, etc.
not_specified = template.Content()

# Escape text for attribute values
def escape(text, quote=None):
    text = str(text)
    text = string.replace(text, '&', '&amp;')
    text = string.replace(text, '<', '&lt;')
    text = string.replace(text, '>', '&gt;',)
    text = string.replace(text, '"', '&quot;')
    return text

# Implement the expr attribute behaviour
class ExprAttrMixin:
    def compile_expr(self, kind = 'eval'):
        if self.has_attrib('expr'):
            self.expr = compile(self.get_attrib('expr'), '<albatross>', kind)
        else:
            self.raise_error('missing "expr" attribute')

    def eval_expr(self, ctx):
        return ctx.eval_expr(self.expr)

# ------------------------------------------------------------------

# The HTML <INPUT> tag is wrapped so that variable values can be
# automatically pulled out of the execution context and rendered as
# the VALUE attribute.  When using Albatross application objects, the
# corresponding browser request is automatically merged back into the
# execution context.
class Input(template.EmptyTag, ExprAttrMixin):
    name = 'al-input'

    def __init__(self, filename, line_num, attribs):
        template.EmptyTag.__init__(self, filename, line_num, attribs)
        attr = string.lower(self.get_attrib('type', 'text'))
        self.unbound_to_html = self.type_dict.get(attr)
        if not self.unbound_to_html:
            self.raise_error('unrecognised <INPUT TYPE="%s">' % attr)
        if self.has_attrib('expr'):
            self.compile_expr()
        for name in ('prevpage', 'nextpage'):
            if self.has_attrib(name):
                self.set_attrib('name',
                                '%s.%s' % (self.get_attrib(name), name))

    def write_attribs_except(self, ctx, *attribs):
        apply(template.EmptyTag.write_attribs_except,
              (self, ctx,
               'name', 'nameexpr',
               'expr', 'valueexpr',
               'nextpage', 'prevpage') + attribs)

    def get_name(self, ctx):
        nameexpr = self.get_attrib('nameexpr')
        if nameexpr:
            return ctx.eval_expr(nameexpr)
        else:
            return self.get_attrib('name')

    def get_name_and_value(self, ctx):
        name = self.get_name(ctx)
        if self.has_attrib('expr'):
            return name, self.eval_expr(ctx)
        else:
            return name, ctx.get_value(name)

    def get_check_value(self, ctx, default = None):
        valueexpr = self.get_attrib('valueexpr')
        if valueexpr:
            return ctx.eval_expr(valueexpr)
        value = ctx.get_value('value')
        if value is None:
            return default
        return value

    def to_html(self, ctx):
        ctx.write_content('<input')
        self.unbound_to_html(self, ctx)
        ctx.write_content('>')

    def text_to_html(self, ctx):
        self.write_attribs_except(ctx, 'value')
        name, value = self.get_name_and_value(ctx)
        ctx.write_content(' name="%s"' % name)
        if value:
            ctx.write_content(' value="%s"' % escape(value))
        ctx.input_add('text', name, value)

    def password_to_html(self, ctx):
        self.write_attribs_except(ctx)
        name = self.get_name(ctx)
        ctx.write_content(' name="%s"' % name)
        ctx.input_add('password', name)

    def radio_to_html(self, ctx):
        self.write_attribs_except(ctx, 'checked')
        name, value = self.get_name_and_value(ctx)
        ctx.write_content(' name="%s"' % name)
        value_attr = self.get_check_value(ctx)
        if not self.has_attrib('value'):
            ctx.write_content(' value="%s"' % value_attr)
        if str(value) == value_attr:
            ctx.write_content(' checked')
        ctx.input_add('radio', name, value_attr)

    def checkbox_to_html(self, ctx):
        self.write_attribs_except(ctx, 'checked')
        name, value = self.get_name_and_value(ctx)
        ctx.write_content(' name="%s"' % name)
        value_attr = self.get_check_value(ctx, 'on')
        if not self.has_attrib('value'):
            ctx.write_content(' value="%s"' % value_attr)
        if type(value) in (type([]), type(())):
            if value_attr in map(str, value):
                ctx.write_content(' checked')
        elif str(value) and value == value_attr:
            ctx.write_content(' checked')
        ctx.input_add('checkbox', name, value_attr)

    def submit_to_html(self, ctx):
        self.write_attribs_except(ctx)
        name = self.get_name(ctx)
        ctx.write_content(' name="%s"' % name)
        ctx.input_add('submit', name)

    def reset_to_html(self, ctx):
        name = self.get_name(ctx)
        ctx.write_content(' name="%s"' % name)
        self.write_attribs_except(ctx)

    def image_to_html(self, ctx):
        self.write_attribs_except(ctx)
        name = self.get_name(ctx)
        ctx.write_content(' name="%s"' % name)
        ctx.input_add('image', name)

    def file_to_html(self, ctx):
        self.write_attribs_except(ctx)
        name = self.get_name(ctx)
        ctx.write_content(' name="%s"' % name)
        ctx.input_add('file', name)

    def hidden_to_html(self, ctx):
        self.write_attribs_except(ctx, 'value')
        name, value = self.get_name_and_value(ctx)
        ctx.write_content(' name="%s"' % name)
        if value:
            ctx.write_content(' value="%s"' % escape(value))
        ctx.input_add('hidden', name, value)

    def button_to_html(self, ctx):
        self.write_attribs_except(ctx)
        name = self.get_name(ctx)
        ctx.write_content(' name="%s"' % name)
        ctx.input_add('button', name)

    type_dict = {
        'text': text_to_html, 'password': password_to_html,
        'radio': radio_to_html, 'checkbox': checkbox_to_html,
        'submit': submit_to_html, 'reset': reset_to_html,
        'image': image_to_html, 'file': file_to_html,
        'hidden': hidden_to_html, 'button': button_to_html, }

# Allows the application need to modify the HREF attribute of the <A>
# tag to include extra information from the execution context.
class Href(template.EnclosingTag, ExprAttrMixin):
    name = 'al-a'

    def __init__(self, filename, line_num, attribs):
        template.EnclosingTag.__init__(self, filename, line_num, attribs)
        if self.has_attrib('expr'):
            self.compile_expr()

    def to_html(self, ctx):
        ctx.write_content('<a')
        self.write_attribs_except(ctx, 'expr', 'href', 'nextpage', 'prevpage')
        href = None
        for name in ('prevpage', 'nextpage'):
            if self.has_attrib(name):
                href = '%s?%s.%s=1' % (ctx.base_url(),
                                       self.get_attrib(name), name)
                break
        if not href:
            if self.has_attrib('expr'):
                href = ctx.eval_expr(self.expr)
            else:
                href = self.get_attrib('href')
            pos = string.find(href, '?')
            if pos < 0 or string.find(href, '=') < pos:
                # need to prepend program URL
                href = '%s?%s' % (ctx.base_url(), href)
        ctx.write_content(' href="%s"' % href)
        ctx.write_content('>')
        template.EnclosingTag.to_html(self, ctx)
        ctx.write_content('</a>')

# Allows the application need to modify the SRC attribute of the <IMG>
# tag to include extra information from the execution context.
class Img(template.EmptyTag, ExprAttrMixin):
    name = 'al-img'

    def __init__(self, filename, line_num, attribs):
        template.EmptyTag.__init__(self, filename, line_num, attribs)
        self.compile_expr()

    def to_html(self, ctx):
        ctx.write_content('<img')
        self.write_attribs_except(ctx, 'expr', 'src')
        ctx.write_content(' src="%s"' % escape(ctx.eval_expr(self.expr)))
        ctx.write_content('>')

# Using the albatross <FORM> tag allows Albatross to record the
# contents of each form.  This is used to register valid browser
# requests with the toolkit.
class Form(template.EnclosingTag):
    name = 'al-form'

    def to_html(self, ctx):
        ctx.write_content('<form')
        self.write_attribs_except(ctx)
        if not self.has_attrib('action'):
            ctx.write_content(' action="%s"' % ctx.base_url())
        ctx.write_content('>')
        ctx.form_open()
        template.EnclosingTag.to_html(self, ctx)
        ctx.form_close()
        ctx.write_content('</form>')

# Applications which need to know the list of valid <al-select>
# options will need to use <al-option> tags in template files.  The
# alternative is to define the option list in Python code.
# 
# Two forms implemented:
# <al-option>12</al-option>
# <al-option value="12">Arbitrary text</al-option>
class Option(template.EnclosingTag):
    name = 'al-option'

    def __init__(self, filename, line_num, attribs):
        template.EnclosingTag.__init__(self, filename, line_num, attribs)

    def to_html(self, ctx, select_value):
        if self.has_attrib('value'):
            value = self.get_attrib('value')
        else:
            ctx.push_content_trap()
            template.EnclosingTag.to_html(self, ctx)
            value = ctx.pop_content_trap()
        ctx.write_content('<option')
        self.write_attribs_except(ctx)
        if type(select_value) is type(''):
            if value == select_value:
                ctx.write_content(' selected')
        elif type(select_value) in (type([]), type(())):
            if value in select_value:
                ctx.write_content(' selected')
        ctx.write_content('>')
        if self.has_attrib('value'):
            template.EnclosingTag.to_html(self, ctx)
        else:
            ctx.write_content(value)

# Some <al-select> tags have their option list supplied by an
# application value.  There are two forms:
# <al-select name="var">
#  <al-option>12</al-option>
# </al-select>
# <al-select name="var" optionexpr="range(12)"/>
class Select(template.EnclosingTag):
    name = 'al-select'

    def __init__(self, filename, line_num, attribs):
        template.EnclosingTag.__init__(self, filename, line_num, attribs)
        if self.has_attrib('optionexpr'):
            self.option_expr = compile(self.get_attrib('optionexpr'), '<albatross>', 'eval')
        if not self.has_attrib('name'):
            self.raise_error('missing "name" attribute')
        self.options = []

    def to_html(self, ctx):
        ctx.write_content('<select')
        self.write_attribs_except(ctx, 'optionexpr')
        ctx.write_content('>')
        value = ctx.get_value(self.get_attrib('name'))
        if type(value) in (type([]), type(())):
            value = map(str, value)
        else:
            value = str(value)
        if self.has_attrib('optionexpr'):
            seq = ctx.eval_expr(self.option_expr)
            for item in seq:
                self.option_to_html(ctx, item, value)
        else:
            for item in self.options:
                item.to_html(ctx, value)
        ctx.write_content('</select>')

    def option_to_html(self, ctx, item, select_value):
        if type(item) is type(()):
            value, text = map(str, item)
        else:
            value = str(item)
            text = value
        ctx.write_content('<option')
        if type(select_value) is type(''):
            if value == select_value:
                ctx.write_content(' selected')
        elif type(select_value) in (type([]), type(())):
            if value in select_value:
                ctx.write_content(' selected')
        if type(item) is type(()):
            ctx.write_content(' value="%s"' % escape(value))
        ctx.write_content('>')
        ctx.write_content(text)
        ctx.write_content('\n')

    def append(self, tag):
        if isinstance(tag, Option):
            self.options.append(tag)
        else:
            template.EnclosingTag.append(self, tag)

class TextArea(template.EnclosingTag):
    name = 'al-textarea'

    def __init__(self, filename, line_num, attribs):
        template.EnclosingTag.__init__(self, filename, line_num, attribs)
        if not self.has_attrib('name'):
            self.raise_error('missing "name" attribute')

    def to_html(self, ctx):
        ctx.write_content('<textarea')
        self.write_attribs_except(ctx)
        ctx.write_content('>')
        name = self.get_attrib('name')
        if ctx.has_value(name):
            value = ctx.get_value(name)
            if not value:
                ctx.write_content('')
            else:
                ctx.write_content(str(value))
        else:
            template.EnclosingTag.to_html(self, ctx)
        ctx.write_content('</textarea>')
        ctx.input_add('textarea', name)

# ------------------------------------------------------------------

# Send HTML accumulated in execution context to output.
class Flush(template.EmptyTag):
    name = 'al-flush'

    def to_html(self, ctx):
        ctx.flush_content()

# Include a template file

# Include a template file. Can name the file directly, or can use an
# expression to yield a filename.
class Include(template.EmptyTag, ExprAttrMixin):
    name = 'al-include'

    def __init__(self, filename, line_num, attribs):
        template.EmptyTag.__init__(self, filename, line_num, attribs)
        if self.has_attrib('expr'):
            self.compile_expr()
        elif self.has_attrib('name'):
            pass
        else:
            self.raise_error('missing "name" or "expr" attribute')

    def to_html(self, ctx):
        if self.has_attrib('expr'):
            name = self.eval_expr(ctx)
        else:
            name = self.get_attrib('name')
        templ = ctx.load_template(name)
        templ.to_html(ctx)

# Conditional content - emulates Python
#     if expr:
#     elif expr:
#     else:
#
# <al-if expr="value > 12">
#  Value > 12
# <al-elif expr="value > 5">
#  Value > 5 and <= 12
# <al-else>
#  Value <= 5
# </al-if>
class Elif(template.EmptyTag, ExprAttrMixin):
    name = 'al-elif'

    def __init__(self, filename, line_num, attribs):
        template.EmptyTag.__init__(self, filename, line_num, attribs)
        self.compile_expr()

    def is_true(self, ctx):
        return self.eval_expr(ctx)

class Else(template.EmptyTag):
    name = 'al-else'

class If(template.EnclosingTag, ExprAttrMixin):
    name = 'al-if'

    def __init__(self, filename, line_num, attribs):
        template.EnclosingTag.__init__(self, filename, line_num, attribs)
        self.compile_expr()
        self.elif_list = []
        self.else_content = not_specified
        self.current_content = self.content

    def to_html(self, ctx):
        if ctx.eval_expr(self.expr):
            template.EnclosingTag.to_html(self, ctx)
            return
        for tag, content in self.elif_list:
            if tag.is_true(ctx):
                content.to_html(ctx)
                return
        self.else_content.to_html(ctx)

    def append(self, tag):
        if isinstance(tag, Elif):
            self.current_content = template.Content()
            self.elif_list.append((tag, self.current_content))
        elif isinstance(tag, Else):
            self.current_content = template.Content()
            self.else_content = self.current_content
        else:
            self.current_content.append(tag)

# Execute some Python code
#
# <al-exec expr="keys = os.environ.keys(); keys.sort()">
class Exec(template.EmptyTag, ExprAttrMixin):
    name = 'al-exec'

    def __init__(self, filename, line_num, attribs):
        template.EmptyTag.__init__(self, filename, line_num, attribs)
        self.compile_expr('exec')

    def to_html(self, ctx):
        ctx.eval_expr(self.expr)

# Iterated content - emulates Python
#     for iter in sequence:
class ListIteratorError(Exception):
    pass

class ListIterator:
    def __init__(self):
        self._index = self._start = self._count = 0
        self._seq = self._have_value = None
        self._pagesize = 0
        # nextpage/prevpage are set by browser request to control
        # pagination.  Once used they are set back to None.  This
        # ensures that the requests are only processed once.
        self.nextpage = self.prevpage = None

    def __getstate__(self):
        return self._start

    def __setstate__(self, start):
        self._start = self._index = start
        self._pagesize = None
        self.nextpage = self.prevpage = self._have_value = self._seq = None

    def __len__(self):
        if self._pagesize:
            return self._pagesize
        elif self._seq is not None:
            return len(self._seq)
        else:
            return 0

    # ----------------------------------------------------
    # Pagination functionality.

    def pagesize(self):
        return self._pagesize

    def set_pagesize(self, size):
        self._pagesize = size

    def handle_pagechange(self):
        if self.nextpage:
            self._start = self._start + self._pagesize
            self.nextpage = None
            self._index = self._start
        elif self.prevpage:
            self._start = self._start - self._pagesize
            if self._start < 0:
                self._start = 0
            self.prevpage = None
            self._index = self._start

    def has_prevpage(self):
        return self._start > 0

    def has_nextpage(self):
        if not self._pagesize:
            raise ListIteratorError('iterator not in page mode')
        return len(self._seq) > self._start + self._pagesize

    # ----------------------------------------------------
    # Sequence control.

    def has_sequence(self):
        return self._seq is not None

    def set_sequence(self, seq):
        self._seq = seq

    # ----------------------------------------------------
    # Iteration.  Make sure that each element is only retrieved from
    # the sequence once.  This allows programs to make use of objects
    # which are not really sequences; popen().readline() for instance.

    def reset_index(self):
        self._index = self._start

    def reset_count(self):
        self._count = 0

    def index(self):
        return self._index

    def count(self):
        return self._count

    def start(self):
        return self._start

    def value(self):
        return self._value

    def clear_value(self):
        self._have_value = None

    def has_value(self):
        if self._have_value is None:
            try:
                self._value = self._seq[self._index]
                self._have_value = 1
            except IndexError:
                self._have_value = 0
        return self._have_value

    def set_value(self, value):
        self._value = value
        self._have_value = 1

    def next(self):
        self._index = self._index + 1
        self._count = self._count + 1
        self._have_value = None
        self.has_value()

# EXPR:seq
#   Yields an object which implements sequence protocol.  The result
#   is assigned to the ITER attribute.  If not present, the ITER
#   attribute must already have a sequence in it.
# PAGESIZE:int
#   Turns on pagination.  If this attribute is used anywhere with the
#   named ITER then that page size becomes the default for all other
#   uses of the ITER.  Places the ITER into the session.
# CONTINUE:bool
#   Instructs the tag not to reset the index() to the start() - it is
#   continued on from the previous ITER use.
# PREPARE:bool
#   Tells the tag to attach the sequence to the ITER and perform
#   pagination calculations.  This makes the ITER available for tests
#   before the content is rendered.
# COLS:int
#   Instructs the tag to render the content in multiple columns.
#   Items are arranged to flow down columns by default.
# FLOW:across/down
#   Define the order in which multicolumn output is arranged.
class For(template.EnclosingTag, ExprAttrMixin):
    name = 'al-for'

    def __init__(self, filename, line_num, attribs):
        template.EnclosingTag.__init__(self, filename, line_num, attribs)
        if self.has_attrib('expr'):
            self.compile_expr()
        if not self.has_attrib('iter'):
            self.raise_error('missing "iter" attribute')
        if self.has_attrib('flow'):
            if string.lower(self.get_attrib('flow')) not in ('across', 'down'):
                self.raise_error('"flow" must be "across" or "down"')

    def to_html(self, ctx):
        # If PAGESIZE or CONTINUE defined then the iterator will not
        # be recreated if it already exists.
        name = self.get_attrib('iter')
        iter = None
        if self.has_attrib('pagesize') or self.has_attrib('continue'):
            iter = ctx.get_value(name)
        if iter is None:
            iter = ListIterator()
            ctx.set_value(name, iter)
        # If iterator does not have a sequence yet, the EXPR attribute
        # must be present
        if self.has_attrib('expr'):
            seq = ctx.eval_expr(self.expr)
            iter.set_sequence(seq)
        elif not iter.has_sequence():
            self.raise_error('no sequence present; expected "expr" attribute')
        # If PAGESIZE is defined then place the iterator into the
        # session and handle prevpage/nextpage events.
        if self.has_attrib('pagesize'):
            ctx.add_session_vars(name)
            pagesize = int(self.get_attrib('pagesize'))
            iter.set_pagesize(pagesize)
            iter.handle_pagechange()
        # If CONTINUE if defined then the contents will flow on from
        # the previous use of the iterator
        if not self.has_attrib('continue'):
            iter.reset_index()
        iter.reset_count()
        iter.has_value()
        # If PREPARE defined then no rendering performed
        if self.has_attrib('prepare'):
            return
        # Render the sequence contents
        if self.has_attrib('cols'):
            if string.lower(self.get_attrib('flow', 'down')) == 'down':
                self.cols_down_html(ctx, iter)
            else:
                self.cols_across_html(ctx, iter)
        else:
            pagesize = iter.pagesize()
            while iter.has_value():
                if pagesize and iter.count() >= pagesize:
                    break
                template.EnclosingTag.to_html(self, ctx)
                iter.next()
            iter.clear_value()

    def cols_down_html(self, ctx, iter):
        num_cols = int(self.get_attrib('cols'))
        num = len(iter)
        num_rows = num / num_cols
        num_extra = num % num_cols
        for row in range(num_rows + (num_extra > 0)):
            index = row
            seq = []
            for col in range(num_cols):
                if row == num_rows and col == num_extra:
                    break
                seq.append(iter._seq[index])
                index = index + num_rows + (col < num_extra)
                if index >= num:
                    break
            iter.set_value(seq)
            template.EnclosingTag.to_html(self, ctx)

    def cols_across_html(self, ctx, iter):
        num_cols = int(self.get_attrib('cols'))
        pagesize = iter.pagesize()
        num = len(iter)
        index = 0
        while index < num:
            seq = iter._seq[index:index + num_cols]
            iter.set_value(seq)
            template.EnclosingTag.to_html(self, ctx)
            index = index + num_cols

class TreeIterator:
    def __init__(self):
        self._tree = self._node = self._have_value = None
        self._stack = []

    def tree_depth(self):
        return self._tree_depth

    def depth(self):
        return len(self._stack)

    def line(self, depth):
        return self._line[depth]

    def value(self):
        return self._value

# Format tree structured data.
#
# <al-tree iter="n" expr="">
#  <al-node name="n">
#  <al-for iter="c" expr="n.depth()">
#   <al-if expr="n.line(c.value()) == 0">  <al-else>
#   <al-if expr="n.line(c.value()) == 1"> |<al-else> \</al-if>
#   </al-if>
#  </al-for>
#  -<al-value expr="n.value().name" whitespace="newline">
#  </al-node>
# </al-tree>
# Would draw something like this:
# -a
#  |-a
#  | |-a
#  | \-b
#  \-b
class Node:
    def __init__(self, child_id):
        self.child_id = child_id

class Tree(template.EnclosingTag, ExprAttrMixin):
    name = 'al-tree'

    def __init__(self, filename, line_num, attribs):
        template.EnclosingTag.__init__(self, filename, line_num, attribs)
        self.compile_expr()
        if not self.has_attrib('iter'):
            self.raise_error('missing "iter" attribute')

    def to_html(self, ctx):
        # Get iterator out of context - allows tree state to be
        # restored from session
        name = self.get_attrib('iter')
        iter = ctx.get_value(name)
        if iter is None:
            iter = TreeIterator()
            ctx.set_value(name, iter)
        tree = ctx.eval_expr(self.expr)
        iter._tree_depth = self.find_tree_depth(tree)
        self.node_to_html(ctx, iter, tree)

    def find_tree_depth(self, tree):
        deepest = 0
        for node in tree.children:
            if not hasattr(node, 'open') or node.open:
                depth = self.find_tree_depth(node)
                if depth > deepest:
                    deepest = depth
        return deepest + 1
        
    def node_to_html(self, ctx, iter, tree):
        parent = None
        line = []
        for node in iter._stack:
            if parent:
                if node is parent.children[-1]:
                    line.append(0)
                else:
                    line.append(1)
            parent = node
        if parent:
            if tree is parent.children[-1]:
                line.append(2)
            else:
                line.append(1)
        iter._line = line
        iter._value = tree
        template.EnclosingTag.to_html(self, ctx)
        iter._stack.append(tree)
        for node in tree.children:
            if not hasattr(node, 'open') or node.open:
                self.node_to_html(ctx, iter, node)
        del iter._stack[-1]

        
# Include execution context values
#
# <al-value expr="name">
# <al-value expr="name" date="%d %b %Y">
# <al-value expr="name" lookup="name_lookup">
#
# When present the DATE attribute specifies a format string which is
# passed to time.strftime() to format the expression value.  The
# LOOKUP value uses the expression value as an index into the named
# lookup table and renders the value as the corresponding table entry.
class Value(template.EmptyTag, ExprAttrMixin):
    name = 'al-value'

    def __init__(self, filename, line_num, attribs):
        template.EmptyTag.__init__(self, filename, line_num, attribs)
        self.compile_expr()

    def to_html(self, ctx):
        value = ctx.eval_expr(self.expr)
        format = self.get_attrib('date')
        if format:
            value = time.strftime(format, time.localtime(value))
            ctx.write_content(value)
            return
        lookup_name = self.get_attrib('lookup')
        if lookup_name:
            lookup = ctx.get_lookup(lookup_name)
            if not lookup:
                self.raise_error('undefined lookup "%s"' % lookup_name)
            lookup.lookup_html(ctx, value)
            return
        ctx.write_content(str(value))

# Specify one entry in a lookup table
#
# <al-item expr="">
#  The enclosed content is evaluated when the table is
#  accessed, not when the table is contructed.  The EXPR
#  attribute is evaluated when the table is evaluated.
#  This allows the application to establish a context in
#  which the item EXPR attributes can be resolved to a
#  constant.
# </al-item>
class Item(template.EnclosingTag, ExprAttrMixin):
    name = 'al-item'

    def __init__(self, filename, line_num, attribs):
        template.EnclosingTag.__init__(self, filename, line_num, attribs)
        self.compile_expr()

# Specify a lookup table for translating a value to arbitrary content
# 
# <al-lookup name="name">
#  <al-item expr="1">Content</al-item>
#  <al-item expr="'hello'">Content</al-item>
#  <al-item expr="value + 3">Content</al-item>
# </al-lookup>
class Lookup(template.EnclosingTag):
    name = 'al-lookup'

    def __init__(self, filename, line_num, attribs):
        template.EnclosingTag.__init__(self, filename, line_num, attribs)
        self.assert_has_attrib('name')
        self.item_list = []

    def append(self, tag):
        if isinstance(tag, Item):
            self.item_list.append(tag)
        else:
            template.EnclosingTag.append(self, tag)

    def to_html(self, ctx):
        ctx.register_lookup(self.get_attrib('name'), self)
        dict = {}
        for item in self.item_list:
            dict[item.eval_expr(ctx)] = item
        self.item_dict = dict

    def lookup_html(self, ctx, value):
        item = self.item_dict.get(value, self.content)
        item.to_html(ctx)

# ------------------------------------------------------------------

# Macro expansions
# <al-macro name="noargs">No args in this one</al-macro>
# <al-macro name="onearg">Put the arg <al-usearg> please</al-macro>
# <al-macro name="namedargs">
#  The unnamed arg can still be used here <al-usearg>
#  But we can also include named arguments <al-usearg name="arg1">
# </al-macro>
#
# Then the macros are expanded like this:
# <al-expand name="noarg"/>
# <al-expand name="onearg>here</al-expand>
# <al-expand name="namedargs">
#  This is the unnamed content
#  <al-setarg name="arg1">This is the named content</al-setarg>
# </al-expand>
#

# The macros are not expanded until they are used in an <al-expand>
# context.  Arguments are placed on a macro execution stack to allow
# macros to be nested within macros and the outer macro inherits
# arguments of the inner macro (and vice-versa).  The content of outer
# macro becomes the unnamed argument of the inner macro.  This sounds
# confusing, but it provides the expected functionality for the user.
# 
# Usearg and Macro do not produce output when evaluated - only when
# expanded via <al-expand>
class Usearg(template.EmptyTag):
    name = 'al-usearg'

    def to_html(self, ctx):
        arg = ctx.get_macro_arg(self.get_attrib('name'))
        arg.to_html(ctx)

class Macro(template.EnclosingTag):
    name = 'al-macro'

    def __init__(self, filename, line_num, attribs):
        template.EnclosingTag.__init__(self, filename, line_num, attribs)
        self.assert_has_attrib('name')

    def to_html(self, ctx):
        ctx.register_macro(self.get_attrib('name'), self)

    def expand_html(self, ctx):
        template.EnclosingTag.to_html(self, ctx)

class Setarg(template.EnclosingTag):
    name = 'al-setarg'

class Expand(template.EnclosingTag):
    name = 'al-expand'

    def __init__(self, filename, line_num, attribs):
        template.EnclosingTag.__init__(self, filename, line_num, attribs)
        self.arg_dict = {}

    def append(self, tag):
        if isinstance(tag, Setarg):
            self.arg_dict[tag.get_attrib('name')] = tag
        else:
            template.EnclosingTag.append(self, tag)

    def to_html(self, ctx):
        dict = {}
        dict[None] = self.content
        dict.update(self.arg_dict)
        ctx.push_macro_args(dict)
        macro = ctx.get_macro(self.get_attrib('name'))
        if not macro:
            self.raise_error('undefined macro "%s"' % self.get_attrib('name'))
        macro.expand_html(ctx)
        ctx.pop_macro_args()

class Call(template.EnclosingTag, ExprAttrMixin):
    name = 'al-call'

    def __init__(self, filename, line_num, attribs):
        template.EnclosingTag.__init__(self, filename, line_num, attribs)
        if self.has_attrib('expr'):
            self.compile_expr('exec')
        self.arg_dict = {}

    def append(self, tag):
        if isinstance(tag, Setarg):
            self.arg_dict[tag.get_attrib('name')] = tag
        else:
            template.EnclosingTag.append(self, tag)

    def to_html(self, ctx):
        dict = {}
        dict[None] = self.content
        dict.update(self.arg_dict)
        ctx.push_macro_args(dict)
        macro = ctx.get_macro(self.get_attrib('name'))
        if not macro:
            self.raise_error('undefined macro "%s"' % self.get_attrib('name'))

        saved_values = {}
        for name in self.get_attrib('params', ()):
            saved_values[name] = ctx.get_value(name)
        if self.has_attrib('expr'):
            ctx.eval_expr(self.expr)
        macro.expand_html(ctx)
        for name in self.get_attrib('params', ()):
            ctx.set_value(name, saved_values[name])

        ctx.pop_macro_args()

tags = (Input, Href, Img, Form, Option, Select, TextArea,
        Flush, Include, Elif, Else, If, Exec, For, Tree,
        Value, Item, Lookup, Usearg, Macro, Setarg, Expand, Call)
